import warnings

import keras
from keras import layers, ops
from keras.utils import Progbar

import tensorflow as tf

import tensorflow_probability as tfp

from . import utils


# Encoder network that processes context points
@keras.saving.register_keras_serializable()
class DeterministicEncoder(layers.Layer):
    def __init__(self, hidden_sizes=[128, 128, 128, 128], **kwargs):
        super().__init__(**kwargs)
        self.hidden_sizes = hidden_sizes
        self.hidden_layers = []
        for size in hidden_sizes[:-1]:
            self.hidden_layers.append(layers.Dense(size, activation="relu"))
        # Final layer without activation
        self.hidden_layers.append(layers.Dense(hidden_sizes[-1]))

    def build(self, input_shape):
        context_x_shape, context_y_shape = input_shape

        # The input to the first Dense layer is the concatenation of x and y features
        current_shape_dim = context_x_shape[-1] + context_y_shape[-1]
        # Build each layer sequentially
        for layer in self.hidden_layers:
            layer.build(input_shape=(None, None, current_shape_dim))
            current_shape_dim = (
                layer.units
            )  # The output dim of this layer is the input dim for the next

        # Mark this layer as built
        super().build(input_shape)

    def call(self, context_x, context_y):
        # Concatenate x and y along feature dimension
        encoder_input = ops.concatenate([context_x, context_y], axis=-1)

        # Pass through MLP
        hidden = encoder_input
        for layer in self.hidden_layers:
            hidden = layer(hidden)

        # Average over context points
        representation = ops.mean(hidden, axis=1, keepdims=True)
        return representation

    def get_config(self):
        config = super().get_config()
        config.update({"hidden_sizes": self.hidden_sizes})
        return config


# Latent encoder network that processes context points for the latent path
@keras.saving.register_keras_serializable()
class LatentEncoder(layers.Layer):
    def __init__(self, hidden_sizes=[128, 128], num_latents=128, **kwargs):
        super().__init__(**kwargs)
        self.hidden_sizes = hidden_sizes
        self.num_latents = num_latents

        self.hidden_layers = []
        for size in hidden_sizes:
            self.hidden_layers.append(layers.Dense(size, activation="relu"))

        # Two final layers for mean and log_sigma
        self.mu_layer = layers.Dense(self.num_latents)
        self.log_sigma_layer = layers.Dense(self.num_latents)

    def build(self, input_shape):
        context_x_shape, context_y_shape = input_shape
        current_shape_dim = context_x_shape[-1] + context_y_shape[-1]

        for layer in self.hidden_layers:
            layer.build(input_shape=(None, None, current_shape_dim))
            current_shape_dim = layer.units

        # The penultimate layer's output is the input to mu and log_sigma layers
        penultimate_shape = (None, None, current_shape_dim)
        self.mu_layer.build(penultimate_shape)
        self.log_sigma_layer.build(penultimate_shape)

        super().build(input_shape)

    def call(self, x, y):
        # Concatenate x and y along feature dimension
        encoder_input = ops.concatenate([x, y], axis=-1)

        # Pass through MLP
        hidden = encoder_input
        for layer in self.hidden_layers:
            hidden = layer(hidden)

        # Aggregate by taking the mean over all points
        hidden = ops.mean(hidden, axis=1)

        # Get mean and log_sigma for the latent distribution
        mu = self.mu_layer(hidden)
        log_sigma = self.log_sigma_layer(hidden)

        # Bound the variance
        sigma = 0.1 + 0.9 * ops.sigmoid(log_sigma)

        # Return a distribution object
        return tfp.distributions.Normal(loc=mu, scale=sigma)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_sizes": self.hidden_sizes,
                "num_latents": self.num_latents,
            }
        )
        return config


# Attentive encoder network that uses self-attention and cross-attention.
@keras.saving.register_keras_serializable()
class AttentiveEncoder(layers.Layer):
    def __init__(self, encoder_sizes=[128, 128, 128, 128], num_heads=8, **kwargs):
        super().__init__(**kwargs)
        self.encoder_sizes = encoder_sizes
        self.num_heads = num_heads
        self.rep_dim = self.encoder_sizes[-1]  # The unified representation dimension

        # --- Sub-layers ---
        # 1. MLP to process each context point (x, y) individually
        self.context_mlp_hidden_layers = []
        for size in self.encoder_sizes[:-1]:
            self.context_mlp_hidden_layers.append(layers.Dense(size, activation="relu"))
        self.context_mlp_final_layer = layers.Dense(self.rep_dim)

        # 2. Self-Attention over context points
        self.self_attention = layers.MultiHeadAttention(
            num_heads=self.num_heads, key_dim=self.rep_dim
        )
        self.self_attention_norm = layers.LayerNormalization()

        # 3. MLP to project target_x to be the cross-attention query
        self.query_mlp = layers.Dense(self.rep_dim)

        # 4. Cross-Attention between target_x and context
        self.cross_attention = layers.MultiHeadAttention(
            num_heads=self.num_heads, key_dim=self.rep_dim
        )
        self.cross_attention_norm = layers.LayerNormalization()

    def build(self, input_shape):
        context_x_shape, context_y_shape, target_x_shape = input_shape

        current_shape_dim = context_x_shape[-1] + context_y_shape[-1]
        for layer in self.context_mlp_hidden_layers:
            layer.build(input_shape=(None, None, current_shape_dim))
            current_shape_dim = layer.units
        self.context_mlp_final_layer.build(input_shape=(None, None, current_shape_dim))

        attention_input_shape = (None, None, self.rep_dim)
        self.self_attention.build(
            query_shape=attention_input_shape, value_shape=attention_input_shape
        )
        self.self_attention_norm.build(attention_input_shape)

        self.query_mlp.build(target_x_shape)

        query_shape = (None, None, self.rep_dim)
        self.cross_attention.build(
            query_shape=query_shape, value_shape=attention_input_shape
        )
        self.cross_attention_norm.build(query_shape)

        super().build(input_shape)

    def call(self, context_x, context_y, target_x):
        encoder_input = ops.concatenate([context_x, context_y], axis=-1)
        hidden = encoder_input
        for layer in self.context_mlp_hidden_layers:
            hidden = layer(hidden)
        context_representation = self.context_mlp_final_layer(hidden)

        self_att_output = self.self_attention(
            query=context_representation,
            value=context_representation,
            key=context_representation,
        )
        context_representation = self.self_attention_norm(
            context_representation + self_att_output
        )

        query = self.query_mlp(target_x)

        cross_att_output = self.cross_attention(
            query=query, value=context_representation, key=context_representation
        )
        final_representation = self.cross_attention_norm(query + cross_att_output)

        return final_representation

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "encoder_sizes": self.encoder_sizes,
                "num_heads": self.num_heads,
            }
        )
        return config


# Decoder network that makes predictions at target points
@keras.saving.register_keras_serializable()
class DeterministicDecoder(layers.Layer):
    def __init__(self, hidden_sizes=[128, 128, 2], **kwargs):
        super().__init__()
        self.hidden_sizes = hidden_sizes

        self.hidden_layers = []
        for size in hidden_sizes[:-1]:
            self.hidden_layers.append(layers.Dense(size, activation="relu"))
        # Final layer outputs mean and log_std
        self.hidden_layers.append(layers.Dense(hidden_sizes[-1]))

    def build(self, input_shape):
        representation_shape, target_x_shape = input_shape

        # The input to the first Dense layer is
        # the concatenation of representation and target_x
        current_shape_dim = representation_shape[-1] + target_x_shape[-1]
        # Build each layer sequentially
        for layer in self.hidden_layers:
            layer.build(input_shape=(None, None, current_shape_dim))
            current_shape_dim = (
                layer.units
            )  # The output dim of this layer is the input dim for the next

        # Mark this layer as built
        super().build(input_shape)

    def call(self, representation, target_x):
        # Expand representation to match target batch size
        # num_targets = ops.shape(target_x)[1]
        # representation = ops.repeat(representation, num_targets, axis=1)

        # Concatenate representation and target_x
        decoder_input = ops.concatenate([representation, target_x], axis=-1)

        # Pass through MLP
        hidden = decoder_input
        for layer in self.hidden_layers:
            hidden = layer(hidden)

        # Split into mean and log_std
        mean, log_std = ops.split(hidden, 2, axis=-1)

        # Bound the variance
        std = 0.1 + 0.9 * ops.softplus(log_std)

        return mean, std

    def get_config(self):
        config = super().get_config()
        config.update({"hidden_sizes": self.hidden_sizes})
        return config


@keras.saving.register_keras_serializable()
class CNP(keras.Model):
    def __init__(
        self,
        encoder_sizes=[128, 128, 128, 128],
        decoder_sizes=[128, 128],
        output_dims=1,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Set encoder and decoder sizes
        self.encoder_sizes = encoder_sizes
        self.decoder_sizes_hidden = list(decoder_sizes)  # Create a mutable
        self.output_dims = output_dims

        full_decoder_sizes = self.decoder_sizes_hidden + [2 * self.output_dims]
        self.encoder = DeterministicEncoder(self.encoder_sizes)
        self.decoder = DeterministicDecoder(full_decoder_sizes)

    def build(self, input_shape):
        context_x_shape, context_y_shape, target_x_shape = input_shape

        # 1. Build the encoder with its specific input shapes.
        #    The encoder's call method expects a tuple of (context_x, context_y).
        self.encoder.build((context_x_shape, context_y_shape))

        # 2. Determine the shape of the representation vector produced by the encoder
        #    to build the decoder. The shape is (batch, 1, last_encoder_size).
        representation_shape = (context_x_shape[0], 1, self.encoder_sizes[-1])

        # 3. Build the decoder with its specific input shapes.
        #    The decoder's call method expects (representation, target_x).
        self.decoder.build((representation_shape, target_x_shape))

        # 4. Run original validation logic from your build method
        xdims = context_x_shape[-1]  # No. of channels / dims in x
        ydims = context_y_shape[-1]  # No. of channels / dims in y
        if xdims > 1:
            warnings.warn(
                "Note: The encoder implementation doesn't implicitly "
                "depend on xdims. Ensure you use an encoder size that is "
                "appropriate for your data.",
                UserWarning,
            )
        assert ydims == self.output_dims

        # 5. Call the parent's build method at the end.
        super().build(input_shape)

    def call(self, inputs, training=False):
        context_x, context_y, target_x = inputs
        representation = self.encoder(
            context_x, context_y
        )  # Get representation from encoder
        mean, std = self.decoder(
            representation, target_x
        )  # Get predictions from decoder
        return mean, std

    @tf.function(jit_compile=True)
    def train_step(self, context_x, context_y, target_x, target_y):
        with tf.GradientTape() as tape:
            mean, std = self(
                (context_x, context_y, target_x), training=True
            )  # Forward pass
            dist = tfp.distributions.MultivariateNormalDiag(mean, std)
            loss_value = -ops.mean(
                dist.log_prob(target_y)
            )  # negative log likelihood loss
        grads = tape.gradient(loss_value, self.trainable_weights)
        self.optimizer.apply(grads, self.trainable_weights)
        return loss_value

    @tf.function(jit_compile=True)
    def test_step(self, context_x, context_y, pred_x):
        # Get model predictions and uncertainty estimates
        pred_y_mean, pred_y_std = self((context_x, context_y, pred_x))
        return pred_y_mean, pred_y_std

    def train(
        self,
        X_train,
        y_train,
        epochs,
        optimizer,
        batch_size=64,
        num_context_range=[2, 10],
        num_context_mode="each",
        dataset_type="gp",
        X_val=None,
        y_val=None,
        plotcb=True,
        pbar=True,
        pred_points=None,
        plot_every=1000,
        seed_val=None,
    ):
        self.optimizer = optimizer

        if pbar:
            progbar = Progbar(epochs, stateful_metrics=["loss"])

        #
        #        if X_val is None:
        #            assert pred_points is not None
        #        elif pred_points is None:
        #            pred_points = X_val.shape[1]
        #
        #        if plotcb:
        #            assert X_val is not None and y_val is not None
        #
        if plotcb:
            assert (
                X_val is not None and y_val is not None
            ), "Validation data must be provided if plotcb is True."
            if pred_points is None and dataset_type == "gp":
                pred_points = X_val.shape[1]

        # Initialize callbacks
        history = keras.callbacks.History()
        callbacks = keras.callbacks.CallbackList([history], model=self)

        callbacks.on_train_begin()  # Called at the start of training

        # no batching yet - possibly implement later
        for epoch in range(1, epochs + 1):
            # Reset metrics at start of each epoch
            callbacks.on_epoch_begin(epoch)  # Called at start of epoch

            X_train_batch, y_train_batch = utils.get_train_batch(
                X_train, y_train, batch_size=batch_size
            )

            num_context = utils._sample_num_context(num_context_range)

            # sample context set from fixed target set
            context_x_train, context_y_train = utils.get_context_set(
                X_train_batch,
                y_train_batch,
                num_context=num_context,
                num_context_mode=num_context_mode,
            )

            target_x_train = X_train_batch
            target_y_train = y_train_batch

            loss = self.train_step(
                context_x_train, context_y_train, target_x_train, target_y_train
            )

            # Update metrics with both loss and mae if reqd.
            logs = {
                "loss": float(loss),
            }
            callbacks.on_epoch_end(epoch, logs)  # Called at end of epoch

            if pbar:
                progbar.update(epoch, values=[("loss", float(loss))])

            if plotcb and (epoch % plot_every == 0):
                if dataset_type == "gp":
                    utils.gplike_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        pred_points=pred_points,
                        num_context_range=num_context_range,
                        num_context_mode=num_context_mode,
                        epoch=epoch,
                        seed=seed_val,
                    )
                elif dataset_type == "mnist":
                    utils.mnist_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        num_context_range=num_context_range,
                        epoch=epoch,
                        seed=seed_val,
                    )

        return history

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "encoder_sizes": self.encoder_sizes,
                "decoder_sizes": self.decoder_sizes_hidden,
                "output_dims": self.output_dims,
            }
        )
        return config


@keras.saving.register_keras_serializable()
class NP(keras.Model):
    def __init__(
        self,
        det_encoder_sizes=[128] * 4,
        latent_encoder_sizes=[128] * 2,
        num_latents=128,
        decoder_sizes=[128] * 2,
        output_dims=1,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.output_dims = output_dims
        self.num_latents = num_latents

        # Store architecture sizes for saving/loading
        self.det_encoder_sizes = det_encoder_sizes
        self.latent_encoder_sizes = latent_encoder_sizes
        self.decoder_sizes_hidden = list(decoder_sizes)

        # Instantiate the components
        self.det_encoder = DeterministicEncoder(self.det_encoder_sizes)
        self.latent_encoder = LatentEncoder(self.latent_encoder_sizes, self.num_latents)

        full_decoder_sizes = self.decoder_sizes_hidden + [2 * self.output_dims]
        # The decoder now needs to handle the concatenated
        # deterministic and latent representations
        self.decoder = DeterministicDecoder(full_decoder_sizes)

    def build(self, input_shape):
        context_x_shape, context_y_shape, target_x_shape, target_y_shape = input_shape
        # 1. Build the deterministic and latent encoders with its specific input shapes.
        self.det_encoder.build((context_x_shape, context_y_shape))
        self.latent_encoder.build((context_x_shape, context_y_shape))

        # 2. Determine the shape of the representation vector produced by the encoder
        # Note that decoder input is concatenation of det_rep and latent_rep
        det_rep_dim = self.det_encoder_sizes[-1]
        latent_rep_dim = self.num_latents
        representation_shape = (None, 1, det_rep_dim + latent_rep_dim)

        # 3. Build the decoder with its specific input shapes.
        self.decoder.build((representation_shape, target_x_shape))

        # 4. Check for shapes
        xdims = context_x_shape[-1]  # No. of channels / dims in x
        ydims = context_y_shape[-1]  # No. of channels / dims in y
        if xdims > 1:
            warnings.warn(
                "Note: The encoder implementation doesn't implicitly "
                "depend on xdims. Ensure you use an encoder size that is "
                "appropriate for your data.",
                UserWarning,
            )
        assert ydims == self.output_dims

        # 5. Call the parent's build method at the end.
        super().build(input_shape)

    def call(self, inputs, training=False):
        context_x, context_y, target_x, target_y = inputs

        # --- Latent Path ---
        # The prior is based on the context set C
        prior_dist = self.latent_encoder(context_x, context_y)

        if training:
            # The posterior is based on the full target set T (which includes C)
            posterior_dist = self.latent_encoder(target_x, target_y)
            # Sample z from the posterior for reconstruction during training
            z = posterior_dist.sample()
        else:
            # At test time, sample from prior as posterior is unavailable
            posterior_dist = None  # Not available at test time
            z = prior_dist.sample()

        # Expand z to match the number of target points
        num_targets = ops.shape(target_x)[1]
        z_rep = ops.repeat(ops.expand_dims(z, axis=1), num_targets, axis=1)

        # --- Deterministic Path ---
        det_rep = self.det_encoder(context_x, context_y)  # Shape: (batch, 1, dim)
        det_rep = ops.repeat(
            det_rep, num_targets, axis=1
        )  # Shape: (batch, num_targets, dim)

        # --- Combine and Decode ---
        representation = ops.concatenate([det_rep, z_rep], axis=-1)
        mean, std = self.decoder(representation, target_x)

        pred_dist = tfp.distributions.MultivariateNormalDiag(mean, std)

        return pred_dist, prior_dist, posterior_dist

    @tf.function(jit_compile=True)
    def train_step(self, context_x, context_y, target_x, target_y):
        with tf.GradientTape() as tape:
            # Forward pass - `training=True` to get posterior
            pred_dist, prior, posterior = self(
                (context_x, context_y, target_x, target_y), training=True
            )

            # 1. Reconstruction Loss
            log_likelihood = pred_dist.log_prob(target_y)

            # 2. KL Divergence (Regularization)
            kl_div = tfp.distributions.kl_divergence(posterior, prior)
            # Sum KL over latent dims, but average over batch
            kl_div = ops.mean(ops.sum(kl_div, axis=-1))

            # 3. ELBO Loss
            # We want to maximize ELBO = log_likelihood - kl_div
            # So we minimize -ELBO = -log_likelihood + kl_div
            num_targets_float = ops.cast(ops.shape(target_x)[1], "float32")
            # The KL is weighted, a common practice to balance the terms
            reconstruction_loss = -ops.mean(log_likelihood)
            loss_value = reconstruction_loss + kl_div / num_targets_float

        grads = tape.gradient(loss_value, self.trainable_weights)
        self.optimizer.apply(grads, self.trainable_weights)

        return loss_value, reconstruction_loss, kl_div

    @tf.function(jit_compile=True)
    def test_step(self, context_x, context_y, pred_x):
        # We don't have target_y, so it's None. `call` will sample from the prior.
        target_y = None
        pred_dist, _, _ = self((context_x, context_y, pred_x, target_y), training=False)
        return pred_dist.mean(), pred_dist.stddev()

    def train(
        self,
        X_train,
        y_train,
        epochs,
        optimizer,
        batch_size=64,
        num_context_range=[2, 10],
        num_context_mode="each",
        dataset_type="gp",
        X_val=None,
        y_val=None,
        plotcb=True,
        pbar=True,
        pred_points=None,
        plot_every=1000,
        seed_val=None,
    ):
        self.optimizer = optimizer

        if pbar:
            progbar = Progbar(epochs, stateful_metrics=["loss", "recon_loss", "kl_div"])

        if plotcb:
            assert (
                X_val is not None and y_val is not None
            ), "Validation data must be provided if plotcb is True."
            if pred_points is None and dataset_type == "gp":
                pred_points = X_val.shape[1]

        # Initialize callbacks
        history = keras.callbacks.History()
        callbacks = keras.callbacks.CallbackList([history], model=self)

        callbacks.on_train_begin()  # Called at the start of training

        # no batching yet - possibly implement later
        for epoch in range(1, epochs + 1):
            # Reset metrics at start of each epoch
            callbacks.on_epoch_begin(epoch)  # Called at start of epoch

            X_train_batch, y_train_batch = utils.get_train_batch(
                X_train, y_train, batch_size=batch_size
            )

            num_context = utils._sample_num_context(num_context_range)

            # sample context set from fixed target set
            context_x_train, context_y_train = utils.get_context_set(
                X_train_batch,
                y_train_batch,
                num_context=num_context,
                num_context_mode=num_context_mode,
            )

            target_x_train = X_train_batch
            target_y_train = y_train_batch

            loss, reconstruction_loss, kl_div = self.train_step(
                context_x_train, context_y_train, target_x_train, target_y_train
            )

            # Update metrics with both loss and mae if reqd.
            logs = {
                "loss": float(loss),
                "reconstruction_loss": float(reconstruction_loss),
                "kl_div": float(kl_div),
            }
            callbacks.on_epoch_end(epoch, logs)  # Called at end of epoch

            if pbar:
                progbar.update(epoch, values=[(k, float(v)) for k, v in logs.items()])

            if plotcb and (epoch % plot_every == 0):
                if dataset_type == "gp":
                    utils.gplike_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        pred_points=pred_points,
                        num_context_range=num_context_range,
                        num_context_mode=num_context_mode,
                        epoch=epoch,
                        seed=seed_val,
                    )
                elif dataset_type == "mnist":
                    utils.mnist_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        num_context_range=num_context_range,
                        epoch=epoch,
                        seed=seed_val,
                    )

        return history

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "det_encoder_sizes": self.det_encoder_sizes,
                "latent_encoder_sizes": self.latent_encoder_sizes,
                "num_latents": self.num_latents,
                "decoder_sizes": self.decoder_sizes_hidden,
                "output_dims": self.output_dims,
            }
        )
        return config


@keras.saving.register_keras_serializable()
class ANP(keras.Model):
    def __init__(
        self,
        encoder_sizes=[128] * 4,  # Simplified from att_encoder_sizes
        num_heads=8,
        latent_encoder_sizes=[128] * 2,
        num_latents=128,
        decoder_sizes=[128] * 2,
        output_dims=1,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.output_dims = output_dims
        self.num_latents = num_latents
        self.num_heads = num_heads

        self.encoder_sizes = encoder_sizes
        self.latent_encoder_sizes = latent_encoder_sizes
        self.decoder_sizes_hidden = list(decoder_sizes)

        self.att_encoder = AttentiveEncoder(self.encoder_sizes, self.num_heads)
        self.latent_encoder = LatentEncoder(self.latent_encoder_sizes, self.num_latents)

        full_decoder_sizes = self.decoder_sizes_hidden + [2 * self.output_dims]
        self.decoder = DeterministicDecoder(full_decoder_sizes)

    def build(self, input_shape):
        (context_x_shape, context_y_shape, target_x_shape, target_y_shape) = input_shape

        self.att_encoder.build((context_x_shape, context_y_shape, target_x_shape))
        self.latent_encoder.build((context_x_shape, context_y_shape))

        det_rep_dim = self.encoder_sizes[-1]
        latent_rep_dim = self.num_latents
        representation_shape = (None, None, det_rep_dim + latent_rep_dim)

        self.decoder.build((representation_shape, target_x_shape))

        super().build(input_shape)

    def call(self, inputs, training=False):
        context_x, context_y, target_x, target_y = inputs

        # --- Latent Path ---
        # The prior is based on the context set C
        prior_dist = self.latent_encoder(context_x, context_y)

        if training:
            # The posterior is based on the full target set T (which includes C)
            posterior_dist = self.latent_encoder(target_x, target_y)
            # Sample z from the posterior for reconstruction during training
            z = posterior_dist.sample()
        else:
            # At test time, sample from prior as posterior is unavailable
            posterior_dist = None  # Not available at test time
            z = prior_dist.sample()

        # Expand z to match the number of target points
        num_targets = ops.shape(target_x)[1]
        z_rep = ops.repeat(ops.expand_dims(z, axis=1), num_targets, axis=1)

        # --- Deterministic Path ---
        det_rep = self.att_encoder(context_x, context_y, target_x)

        # --- Combine and Decode ---
        representation = ops.concatenate([det_rep, z_rep], axis=-1)
        mean, std = self.decoder(representation, target_x)

        pred_dist = tfp.distributions.MultivariateNormalDiag(mean, std)

        return pred_dist, prior_dist, posterior_dist

    @tf.function(jit_compile=True)
    def train_step(self, context_x, context_y, target_x, target_y):
        with tf.GradientTape() as tape:
            # Forward pass - `training=True` to get posterior
            pred_dist, prior, posterior = self(
                (context_x, context_y, target_x, target_y), training=True
            )

            # 1. Reconstruction Loss
            log_likelihood = pred_dist.log_prob(target_y)

            # 2. KL Divergence (Regularization)
            kl_div = tfp.distributions.kl_divergence(posterior, prior)
            # Sum KL over latent dims, but average over batch
            kl_div = ops.mean(ops.sum(kl_div, axis=-1))

            # 3. ELBO Loss
            # We want to maximize ELBO = log_likelihood - kl_div
            # So we minimize -ELBO = -log_likelihood + kl_div
            num_targets_float = ops.cast(ops.shape(target_x)[1], "float32")
            # The KL is weighted, a common practice to balance the terms
            reconstruction_loss = -ops.mean(log_likelihood)
            loss_value = reconstruction_loss + kl_div / num_targets_float

        grads = tape.gradient(loss_value, self.trainable_weights)
        self.optimizer.apply(grads, self.trainable_weights)

        return loss_value, reconstruction_loss, kl_div

    @tf.function(jit_compile=True)
    def test_step(self, context_x, context_y, pred_x):
        # We don't have target_y, so it's None. `call` will sample from the prior.
        target_y = None
        pred_dist, _, _ = self((context_x, context_y, pred_x, target_y), training=False)
        return pred_dist.mean(), pred_dist.stddev()

    def train(
        self,
        X_train,
        y_train,
        epochs,
        optimizer,
        batch_size=64,
        num_context_range=[2, 10],
        num_context_mode="each",
        dataset_type="gp",
        X_val=None,
        y_val=None,
        plotcb=True,
        pbar=True,
        pred_points=None,
        plot_every=1000,
        seed_val=None,
    ):
        self.optimizer = optimizer

        if pbar:
            progbar = Progbar(epochs, stateful_metrics=["loss", "recon_loss", "kl_div"])

        if plotcb:
            assert (
                X_val is not None and y_val is not None
            ), "Validation data must be provided if plotcb is True."
            if pred_points is None and dataset_type == "gp":
                pred_points = X_val.shape[1]

        # Initialize callbacks
        history = keras.callbacks.History()
        callbacks = keras.callbacks.CallbackList([history], model=self)

        callbacks.on_train_begin()  # Called at the start of training

        # no batching yet - possibly implement later
        for epoch in range(1, epochs + 1):
            # Reset metrics at start of each epoch
            callbacks.on_epoch_begin(epoch)  # Called at start of epoch

            X_train_batch, y_train_batch = utils.get_train_batch(
                X_train, y_train, batch_size=batch_size
            )

            num_context = utils._sample_num_context(num_context_range)

            # sample context set from fixed target set
            context_x_train, context_y_train = utils.get_context_set(
                X_train_batch,
                y_train_batch,
                num_context=num_context,
                num_context_mode=num_context_mode,
            )

            target_x_train = X_train_batch
            target_y_train = y_train_batch

            loss, reconstruction_loss, kl_div = self.train_step(
                context_x_train, context_y_train, target_x_train, target_y_train
            )

            # Update metrics with both loss and mae if reqd.
            logs = {
                "loss": float(loss),
                "reconstruction_loss": float(reconstruction_loss),
                "kl_div": float(kl_div),
            }
            callbacks.on_epoch_end(epoch, logs)  # Called at end of epoch

            if pbar:
                progbar.update(epoch, values=[(k, float(v)) for k, v in logs.items()])

            if plotcb and (epoch % plot_every == 0):
                if dataset_type == "gp":
                    utils.gplike_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        pred_points=pred_points,
                        num_context_range=num_context_range,
                        num_context_mode=num_context_mode,
                        epoch=epoch,
                        seed=seed_val,
                    )
                elif dataset_type == "mnist":
                    utils.mnist_val_step(
                        model=self,
                        X_val=X_val,
                        y_val=y_val,
                        num_context_range=num_context_range,
                        epoch=epoch,
                        seed=seed_val,
                    )
        return history

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "encoder_sizes": self.encoder_sizes,
                "num_heads": self.num_heads,
                "latent_encoder_sizes": self.latent_encoder_sizes,
                "num_latents": self.num_latents,
                "decoder_sizes": self.decoder_sizes_hidden,
                "output_dims": self.output_dims,
            }
        )
        return config



import keras_neural_processes as knp


def test_CNP():
    model = knp.CNP()
    assert model


def test_NP():
    model = knp.NP()
    assert model


def test_ANP():
    model = knp.ANP()
    assert model



import warnings

from keras import ops

import matplotlib.pyplot as plt

import numpy as np

import tensorflow as tf

import tensorflow_probability as tfp


def get_context_set(
    target_x,
    target_y,
    num_context,
    # either a single no. (total or each)
    # or a list ([each1, each2 ...eachn])
    # where n=#ofchannels aka np.unique(target_x[:,:,1])
    num_context_mode="all",  # or "each"
    seed=None,
    # ADD SUPPORT FOR RANGE LATER - then sample different num
    # context_frac_range=[0.01, 0.05],
):
    """
    Selects a subset of points from the target set to form the context set.
    Sid note: shape is (samples, points, channels) for x and y both

    Args:
        target_x (np.array or tf.Tensor):
            Shape (batch_size, num_points, x_channels).
            For multi-channel data in 'long' format, x_channels is typically
            2 for (time, channel_id).
        target_y (np.array or tf.Tensor):
            Shape (batch_size, num_points, y_channels).
        num_context (int or list/tuple):
            - If mode is "all": An integer specifying the total number of
            context points.
            - If mode is "each": An integer for sampling that many points from *each*
              channel, OR a list/tuple specifying the number of points for each
              channel individually (e.g., [10, 5] for 2 channels).
        num_context_mode (str):
            - "all": Randomly sample `num_context` points from the entire set,
              regardless of channel.
            - "each": Sample a specified number of points from each channel separately.
        seed (int, optional): Seed for the random number generator.

    """
    assert (
        target_x.shape[:-1] == target_y.shape[:-1]
    )  # the inputs should have same no. of samples and points, channels may differ
    n_points = target_x.shape[1]

    if isinstance(num_context, (list, tuple)) and num_context_mode != "each":
        raise ValueError(
            "For a list-like collection of num_context, the mode must be 'each'."
        )

    # Convert to array if tensor
    target_x_np, target_y_np = ops.convert_to_numpy(target_x), ops.convert_to_numpy(
        target_y
    )
    rng = np.random.default_rng(seed)

    final_indices = []

    if num_context_mode == "all":
        # Sample `num_context` indices from the total number of points.
        final_indices = rng.choice(n_points, num_context, replace=False)
    elif num_context_mode == "each":
        # --- NEW: Vectorized Grouping using argsort ---
        # This is much faster than looping and calling np.where.

        # 1. Get all channel IDs from the first sample (structure is same for all)
        all_channel_ids = target_x_np[0, :, 1]

        # 2. Get the indices that would sort the channel IDs
        sort_idx = np.argsort(all_channel_ids, kind="stable")

        # 3. Find the locations where the channel
        # ID changes. These are the boundaries of each group.
        #    We use np.diff on the sorted channel IDs to find these boundaries.
        sorted_channel_ids = all_channel_ids[sort_idx]
        # Adding a value at the end to ensure the last group is captured
        diff_array = np.diff(np.append(sorted_channel_ids, np.inf))
        split_points = np.where(diff_array)[0] + 1

        # 4. `split` the sorted original indices into groups based on the boundaries.
        #    Each element in this list is an array of original indices for one channel.
        grouped_original_indices = np.split(sort_idx, split_points[:-1])
        # --- End of Vectorized Grouping ---

        channels = np.unique(sorted_channel_ids)
        n_channels = len(channels)

        if isinstance(num_context, (list, tuple)):
            assert (
                len(num_context) == n_channels
            ), f"Length of num_context list ({len(num_context)}) "
            f"must match the number of channels ({n_channels})"

        all_channel_indices = []
        # This loop is now very fast as it just operates on pre-calculated groups
        for i, channel_indices in enumerate(grouped_original_indices):
            if isinstance(num_context, (list, tuple)):
                num_to_sample = num_context[i]
            else:
                num_to_sample = num_context

            if num_to_sample > len(channel_indices):
                raise ValueError(
                    f"Requesting {num_to_sample} context points "
                    f"for channel {channels[i]}, "
                    f"but only {len(channel_indices)} are available."
                )

            chosen_indices = rng.choice(channel_indices, num_to_sample, replace=False)
            all_channel_indices.append(chosen_indices)

        final_indices = np.concatenate(all_channel_indices)

    else:
        raise AttributeError("num_context_mode can only be 'all' or 'each'.")

    context_x = target_x_np[:, final_indices, :]
    context_y = target_y_np[:, final_indices, :]

    if tf.is_tensor(target_x):
        context_x = tf.constant(context_x)
        context_y = tf.constant(context_y)

    return context_x, context_y


def get_train_batch(X_train, y_train, batch_size=64):
    """
    Sidnote: shape is (samples, points, channels) for x and y both
    """
    assert X_train.shape[:-1] == y_train.shape[:-1]
    tot_samples = X_train.shape[0]
    # choose batch_size points from tot_batches e.g. 64 from 10000 randomly
    batch_indices = np.random.choice(tot_samples, batch_size, replace=False)

    if isinstance(X_train, np.ndarray):
        X_train_batch = X_train[batch_indices, :, :]
        y_train_batch = y_train[batch_indices, :, :]
    elif tf.is_tensor(X_train):
        X_train_batch = tf.gather(X_train, batch_indices, axis=0)
        y_train_batch = tf.gather(y_train, batch_indices, axis=0)
    else:
        print("X_train is neither a NumPy array nor a TensorFlow tensor")

    return X_train_batch, y_train_batch


def gplike_calculate_mymetrics(
    pred_x,
    pred_y_mean,
    pred_y_std,
    target_x,
    target_y,
    context_x,
    context_y,
    closeness_thresh=0.1,
    std_multiplier=1,
):
    """
    Calculates metrics directly on long-format data.
    This version is robust to uneven numbers of context points per channel.
    """
    # Convert all to numpy for calculation
    pred_x, pred_y_mean, pred_y_std, target_x, target_y, context_x, context_y = map(
        ops.convert_to_numpy,
        [pred_x, pred_y_mean, pred_y_std, target_x, target_y, context_x, context_y],
    )

    # 1. Target Coverage: Fraction of truth points within 1-sigma uncertainty
    upper_bound = pred_y_mean + pred_y_std
    lower_bound = pred_y_mean - pred_y_std
    target_in_bounds = (target_y >= lower_bound) & (target_y <= upper_bound)
    target_coverage = np.mean(target_in_bounds)

    # 2. Context Reconstruction Coverage
    # For each context point, find the closest prediction and check if it's in bounds.
    batch_size = context_x.shape[0]
    context_coverage_scores = []
    context_is_close_scores = []

    for i in range(batch_size):
        # Find predictions corresponding to context points
        # This is simpler as target_x is the same as the full x space of the function
        # We assume context_x points exist within target_x
        # Create a mapping from a tuple of coords to an index
        target_map = {tuple(row): idx for idx, row in enumerate(target_x[i])}
        context_indices_in_target = [target_map[tuple(row)] for row in context_x[i]]

        pred_mean_at_context = pred_y_mean[i, context_indices_in_target]
        true_val_at_context = target_y[i, context_indices_in_target]
        pred_std_at_context = pred_y_std[i, context_indices_in_target]

        is_context_close = (
            np.abs(true_val_at_context - pred_mean_at_context) <= closeness_thresh
        )
        context_is_close_scores.append(is_context_close)

        upper = pred_mean_at_context + pred_std_at_context
        lower = pred_mean_at_context - pred_std_at_context

        coverage = (context_y[i] >= lower) & (context_y[i] <= upper)
        context_coverage_scores.append(np.mean(coverage))

    return {
        "target_coverage": target_coverage,
        "fraction_context_close": np.mean(context_is_close_scores),
        "context_reconstruction_coverage": np.mean(context_coverage_scores),
        "mean_predictive_confidence": np.mean(pred_y_std),
    }


def gplike_plot_functions(
    ax,
    pred_x,
    pred_y_mean,
    pred_y_std,
    target_x,
    target_y,
    context_x,
    context_y,
    objnum=0,
):
    """
    Plots data in the new [time, channel_id] -> value format.
    It separates the data by channel_id for plotting.
    """
    pred_x, pred_y_mean, pred_y_std, target_x, target_y, context_x, context_y = map(
        ops.convert_to_numpy,
        [pred_x, pred_y_mean, pred_y_std, target_x, target_y, context_x, context_y],
    )
    # Define colors for the channels

    trueval_colors = ["#0072B2", "#D55E00", "#009E73", "#CC79A7", "#F0E442", "#999999"]
    predval_colors = ["#56B4E9", "#E69F00", "#66c5ab", "#e0afca", "#f6ef8e", "#c2c2c2"]
    std_bgcolors = ["#9AC2E0", "#F3C4A5", "#b2e2d5", "#f0d7e5", "#faf7c6", "#e0e0e0"]

    y_channels = np.unique(pred_x[:, :, 1])
    n_y_channels = len(y_channels)
    if n_y_channels > len(trueval_colors):
        warnings.warn(f"Plotting supports a maximum of {len(trueval_colors)} channels.")

    for i, y_channel in enumerate(y_channels):
        if i >= len(trueval_colors):
            break
        ych_num = i + 1

        # Plot truth first
        channel_mask = target_x[objnum, :, 1] == y_channel
        ax.plot(
            target_x[objnum, channel_mask, 0],
            target_y[objnum, channel_mask, 0],
            color=trueval_colors[i],
            linestyle=":",
            linewidth=2,
            label=f"truth$_{{y{ych_num}}}$",
        )

        # Plot context points second
        channel_mask_ctx = context_x[objnum, :, 1] == y_channel
        ax.plot(
            context_x[objnum, channel_mask_ctx, 0],
            context_y[objnum, channel_mask_ctx, 0],
            color=trueval_colors[i],
            linestyle="",
            marker="o",
            markersize=10,
            label=f"context$_{{y{ych_num}}}$",
        )

        # Plot predictions third
        channel_mask_pred = pred_x[objnum, :, 1] == y_channel
        ax.plot(
            pred_x[objnum, channel_mask_pred, 0],
            pred_y_mean[objnum, channel_mask_pred, 0],
            color=predval_colors[i],
            linewidth=2,
            label=f"pred$_{{y{ych_num}}}$",
        )

        ax.fill_between(
            pred_x[objnum, channel_mask_pred, 0],
            pred_y_mean[objnum, channel_mask_pred, 0]
            - pred_y_std[objnum, channel_mask_pred, 0],
            pred_y_mean[objnum, channel_mask_pred, 0]
            + pred_y_std[objnum, channel_mask_pred, 0],
            alpha=0.2,
            facecolor=std_bgcolors[i],
            interpolate=True,
        )
    # ax.set_yticks([-2, 0, 2])
    # ax.set_xticks([-2, 0, 2])
    # ax.set_ylim([-2, 2])
    ax.set_facecolor("white")

    ax.tick_params(axis="both", which="major", labelsize=16)
    ax.grid(True, which="both", linestyle="--", linewidth=0.5)
    return ax


def gplike_fixed_sets(
    X_val, y_val, pred_points, seed, fixed_val_num_context, num_context_mode
):
    # no default seed as very important for valstep choice

    context_x_val_fixed, context_y_val_fixed = get_context_set(
        X_val,
        y_val,
        num_context=fixed_val_num_context,
        seed=seed,
        num_context_mode=num_context_mode,
    )
    target_x_val_fixed, target_y_val_fixed = X_val, y_val
    # pred_x_val_fixed = np.linspace(-2, 2, pred_points).reshape(
    #     target_x_val_fixed.shape[0], pred_points, target_x_val_fixed.shape[2]
    # )
    pred_x_val_fixed = target_x_val_fixed

    return (
        context_x_val_fixed,
        context_y_val_fixed,
        target_x_val_fixed,
        target_y_val_fixed,
        pred_x_val_fixed,
    )


def _is_per_channel_range(num_context_range):
    """Checks if num_context_range is in the per-channel format."""
    return (
        isinstance(num_context_range, (list, tuple))
        and len(num_context_range) > 0
        and all(isinstance(i, (list, tuple)) for i in num_context_range)
    )


def _sample_num_context(num_context_range):
    """
    Samples the number of context points based on the format of num_context_range.

    Args:
        num_context_range (list or list of lists):
            - [min, max]: Samples a single integer from this range.
            - [[min1, max1], [min2, max2], ...]: Samples one integer from each
              sub-list's range and returns a list of integers.

    Returns:
        int or list: The number of context points to sample.
    """
    if _is_per_channel_range(num_context_range):
        # Per-channel format: sample one integer from each sub-list's range.
        return [np.random.randint(low=r[0], high=r[1] + 1) for r in num_context_range]
    else:
        # Simple format: sample a single integer.
        return np.random.randint(
            low=num_context_range[0], high=num_context_range[1] + 1
        )


def _get_fixed_num_context(num_context_range):
    """
    Gets the fixed number of context points (mean of the range)
    based on the format of num_context_range.

    Args:
        num_context_range (list or list of lists):
            - [min, max]: Returns the integer mean of the range.
            - [[min1, max1], [min2, max2], ...]: Returns a list containing the
              integer mean of each sub-list's range.

    Returns:
        int or list: The fixed number of context points for validation.
    """
    if _is_per_channel_range(num_context_range):
        # Per-channel format: return a list of the means.
        return [int(np.mean(r)) for r in num_context_range]
    else:
        # Simple format: return the single mean.
        return int(np.mean(num_context_range))


def gplike_new_sets(X_val, y_val, pred_points, num_context_range, num_context_mode):
    num_context = _sample_num_context(num_context_range)

    context_x_val, context_y_val = get_context_set(
        X_val,
        y_val,
        num_context=num_context,
        num_context_mode=num_context_mode,
    )
    target_x_val, target_y_val = X_val, y_val
    # pred_x_val = np.linspace(-2, 2, pred_points).reshape(
    #     target_x_val.shape[0], pred_points, target_x_val.shape[2]
    # )
    pred_x_val = target_x_val
    return context_x_val, context_y_val, target_x_val, target_y_val, pred_x_val


def gplike_val_step(
    model, X_val, y_val, pred_points, num_context_range, num_context_mode, epoch, seed
):
    # -------- Part 1: Fixed validation set --------

    fixed_val_num_context = _get_fixed_num_context(num_context_range)
    # Get fixed set
    (
        context_x_val_fixed,
        context_y_val_fixed,
        target_x_val_fixed,
        target_y_val_fixed,
        pred_x_val_fixed,
    ) = gplike_fixed_sets(
        X_val,
        y_val,
        pred_points,
        seed,
        fixed_val_num_context=fixed_val_num_context,
        num_context_mode=num_context_mode,
    )

    # Get model predictions and uncertainty estimates
    pred_y_mean_fixed, pred_y_std_fixed = model.test_step(
        context_x_val_fixed, context_y_val_fixed, pred_x_val_fixed
    )

    # Get model predictions and uncertainty estimates
    dist_fixed = tfp.distributions.MultivariateNormalDiag(
        pred_y_mean_fixed, pred_y_std_fixed
    )
    val_loss_fixed = -ops.mean(dist_fixed.log_prob(target_y_val_fixed))
    print(f"\nEpoch{epoch}, val_loss_fixed={val_loss_fixed} (lower is better)")

    # -------- Part 2: Calculate and print goodness
    # of fit metrics for fixed validation set  --------

    mymetrics = gplike_calculate_mymetrics(
        pred_x_val_fixed,
        pred_y_mean_fixed,
        pred_y_std_fixed,
        target_x_val_fixed,
        target_y_val_fixed,
        context_x_val_fixed,
        context_y_val_fixed,
    )

    for metric_name, value in mymetrics.items():
        print(f"{metric_name}: {value:.1%} (higher is better)")

    # -------- Part 3: New random validation set (in addition to fixed set) --------

    # Generate new set
    context_x_val, context_y_val, target_x_val, target_y_val, pred_x_val = (
        gplike_new_sets(
            X_val,
            y_val,
            pred_points,
            num_context_range,
            num_context_mode=num_context_mode,
        )
    )

    # Get model predictions and uncertainty estimates
    pred_y_mean, pred_y_std = model.test_step(context_x_val, context_y_val, pred_x_val)

    # # Calc and print val_loss for non fixed data
    # dist = tfp.distributions.MultivariateNormalDiag(pred_y_mean, pred_y_std)
    # val_loss = -ops.mean(dist.log_prob(target_y_val))
    # print(f"\nEpoch{epoch}, val_loss={val_loss} (lower is better)")

    # -------- Part 4: Plot val predictions for fixed and random set  --------
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))

    gplike_plot_functions(
        ax=ax1,
        pred_x=pred_x_val_fixed,
        pred_y_mean=pred_y_mean_fixed,
        pred_y_std=pred_y_std_fixed,
        target_x=target_x_val_fixed,
        target_y=target_y_val_fixed,
        context_x=context_x_val_fixed,
        context_y=context_y_val_fixed,
    )
    ax1.set_title("Fixed validation set")

    gplike_plot_functions(
        ax=ax2,
        pred_x=pred_x_val,
        pred_y_mean=pred_y_mean_fixed,
        pred_y_std=pred_y_std_fixed,
        target_x=target_x_val,
        target_y=target_y_val,
        context_x=context_x_val,
        context_y=context_y_val,
    )
    ax2.set_title("Random validation set")
    # ax2.legend(bbox_to_anchor=(1.27, 1), loc="upper right")

    # Create a single legend for the figure
    handles, labels = ax2.get_legend_handles_labels()
    # fig.legend(handles, labels, loc="center", bbox_to_anchor=(0.5, 0.5))

    fig.legend(
        handles,
        labels,
        loc="lower center",  # Anchor point is the bottom-center of the legend box
        bbox_to_anchor=(0.5, 0.78),  # Position legend's bottom at y=0.88
        ncol=3,  # Arrange in 3 columns; you can tweak this number
    )

    fig.suptitle(f"Epoch {epoch}")
    plt.tight_layout(rect=[0, 0, 1, 0.88])
    plt.show()
    plt.close()


def plot_2dimage(
    ax,
    image,
    title=None,
    cmap="gray",
    vmin=None,
    vmax=None,
    plot_context=False,
    context_x=None,
):
    # 1. Plot the image
    im = ax.imshow(image, cmap=cmap, vmin=vmin, vmax=vmax, origin="upper")

    # Add a colorbar, which is especially useful for the uncertainty plot
    fig = ax.get_figure()
    fig.colorbar(im, ax=ax, fraction=0.046, pad=0.04)

    # 2. Plot context points if requested
    if plot_context:
        if context_x is None:
            raise ValueError(
                "`context_x` must be provided when `plot_context` is True."
            )

        height, width = image.shape
        # De-normalize context coordinates from [-1, 1] to pixel space [0, 27]
        context_x_pixels = (context_x[0, :, 0] + 1) * (width - 1) / 2
        context_y_pixels = (context_x[0, :, 1] + 1) * (height - 1) / 2

        ax.scatter(
            context_x_pixels,
            context_y_pixels,
            s=40,
            alpha=0.7,
            facecolors="none",
            edgecolors="red",
            linewidth=1.5,
            label="Context",
        )

    # 3. Set plot aesthetics
    if title:
        ax.set_title(title, fontsize=12)
    ax.set_xticks([])
    ax.set_yticks([])


def mnist_val_step(model, X_val, y_val, num_context_range, epoch, seed):
    if _is_per_channel_range(num_context_range):
        raise ValueError(
            "MNIST dataset is single-channel, but num_context_range "
            "was provided in the per-channel format. Please use [min, max]."
        )
    # -------- Part 1: Fixed validation set --------
    fixed_val_num_context = int(np.mean(num_context_range))
    context_x_val, context_y_val = get_context_set(
        X_val, y_val, num_context=fixed_val_num_context, seed=seed
    )
    target_x_val = X_val

    # Get model predictions and uncertainty estimates
    pred_y_mean, pred_y_std = map(
        ops.convert_to_numpy,
        model.test_step(context_x_val, context_y_val, target_x_val),
    )

    # -------- Part 2: Plot fixed validation set --------

    # --- Create Plots ---
    # We now create a 1x3 grid to include the standard deviation plot
    fig, axes = plt.subplots(1, 3, figsize=(15, 5))
    fig.suptitle(f"Epoch {epoch}", fontsize=16)

    # Reshape images from flat vectors to 2D arrays
    # Assuming batch size is 1 for validation plotting
    IMAGE_HEIGHT, IMAGE_WIDTH = 28, 28  # MNIST
    gt_img = y_val[0].reshape(IMAGE_HEIGHT, IMAGE_WIDTH)
    pred_img = pred_y_mean[0].reshape(IMAGE_HEIGHT, IMAGE_WIDTH)
    std_img = pred_y_std[0].reshape(IMAGE_HEIGHT, IMAGE_WIDTH)

    # Plot 1: Ground Truth with Context Points
    plot_2dimage(
        axes[0],
        gt_img,
        "Ground Truth with Context",
        vmin=-0.5,
        vmax=0.5,
        plot_context=True,
        context_x=context_x_val,
    )

    # Plot 2: Prediction (Mean)
    plot_2dimage(axes[1], pred_img, "Prediction (Mean)", vmin=-0.5, vmax=0.5)

    # Plot 3: Uncertainty (Standard Deviation)
    # Using a different colormap like 'viridis' or 'magma' is great for heatmaps
    plot_2dimage(axes[2], std_img, "Uncertainty (Std Dev)", cmap="magma")

    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.show()
    plt.close()



from .src import (
    ANP,
    AttentiveEncoder,
    CNP,
    DeterministicDecoder,
    DeterministicEncoder,
    LatentEncoder,
    NP,
)

# from .utils import (
#     get_context_set,
#     get_train_batch,
#     gplike_calculate_mymetrics,
#     gplike_fixed_sets,
#     gplike_new_sets,
#     gplike_plot_functions,
#     gplike_val_step,
#     mnist_val_step,
#     plot_2dimage,
# )

__version__ = "0.0.1"

__all__ = [
    "get_context_set",
    "get_train_batch",
    "gplike_calculate_mymetrics",
    "gplike_plot_functions",
    "gplike_fixed_sets",
    "gplike_new_sets",
    "gplike_val_step",
    "plot_2dimage",
    "mnist_val_step",
    "DeterministicEncoder",
    "LatentEncoder",
    "AttentiveEncoder",
    "DeterministicDecoder",
    "CNP",
    "NP",
    "ANP",
]
